SuperStrict

Type TVoxelOctree
	Const MAX_SCALE:Int         = 23
	Const EPSILON:Float         = 2^-MAX_SCALE
	Const INIT_DATA_SIZE:Int    = 512*512*4
	Const INIT_POINTER_SIZE:Int = 512*4
	Const ALLOC_STEP:Int        = 512*512*4
	
	Const MISS:Float = 10000.0
	
	Global BitCount:Int[] = [..
		0, 1, 1, 2, 1, 2, 2, 3, 1, 2, 2, 3, 2, 3, 3, 4, ..
		1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5, ..
		1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5, ..
		2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6, ..
		1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5, ..
		2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6, ..
		2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6, ..
		3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7, ..
		1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5, ..
		2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6, ..
		2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6, ..
		3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7, ..
		2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6, ..
		3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7, ..
		3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7, ..
		4, 5, 5, 6, 5, 6, 6, 7, 5, 6, 6, 7, 6, 7, 7, 8 ..
	]
	
	
	Field Data:Int Ptr
	Field DataSize:Int
	
	Field FarPointers:Int Ptr
	Field FarPointerCount:Int
	Field FarPointerSize:Int
	
	Method Raymarch:Float(OX:Float, OY:Float, OZ:Float, DX:Float, DY:Float, DZ:Float, RaySizeCoef:Float, RaySizeBias:Float, Normal:Int Var, MinLevel:Int = 1)
		Local Stack:Int[MAX_SCALE + 1, 2]
		
		OX :+ 1.5
		OY :+ 1.5
		OZ :+ 1.5
		
		If Abs(DX) < EPSILON Then DX = EPSILON
		If Abs(DY) < EPSILON Then DY = EPSILON
		If Abs(DZ) < EPSILON Then DZ = EPSILON
		
		Local dTx:Float = 1.0/(-Abs(DX))
		Local dTy:Float = 1.0/(-Abs(DY))
		Local dTz:Float = 1.0/(-Abs(DZ))
		
		Local bTx:Float = dTx*OX
		Local bTy:Float = dTy*OY
		Local bTz:Float = dTz*OZ
		
		Local OctantMask:Int = %111
		If DX > 0.0 Then OctantMask = OctantMask ~ %001; bTx = 3.0*dTx - bTx
		If DY > 0.0 Then OctantMask = OctantMask ~ %010; bTy = 3.0*dTy - bTy
		If DZ > 0.0 Then OctantMask = OctantMask ~ %100; bTz = 3.0*dTz - bTz
		
		Local MinT:Float = Max(2.0*dTx - bTx, Max(2.0*dTy - bTy, 2.0*dTz - bTz))
		Local MaxT:Float = Min(    dTx - bTx, Min(    dTy - bTy,     dTz - bTz))
		Local H:Float = MaxT
		MinT = Max(MinT, 0.0)
		'MaxT = Min(MaxT, 1.0)
		
		Local Current:Int = 0
		Local Parent:Int  = 0
		Local IDx:Int     = 0
		Local PosX:Float  = 1.0
		Local PosY:Float  = 1.0
		Local PosZ:Float  = 1.0
		Local Scale:Int   = MAX_SCALE - 1
		
		Local Scale_Exp2:Float = 0.5
		
		If 1.5*dTx - bTx > MinT Then IDx = IDx ~ %001; PosX = 1.5
		If 1.5*dTy - bTy > MinT Then IDx = IDx ~ %010; PosY = 1.5
		If 1.5*dTz - bTz > MinT Then IDx = IDx ~ %100; PosZ = 1.5
		
		While Scale < MAX_SCALE
			If Current = 0 Then Current = Data[Parent]
			
			Local CornerTX:Float = PosX*dTx - bTx
			Local CornerTY:Float = PosY*dTy - bTy
			Local CornerTZ:Float = PosZ*dTz - bTz
			Local MaxTC:Float = Min(CornerTX, Min(CornerTY, CornerTZ))
			
			Local ChildShift:Int = IDx ~ OctantMask
			Local ChildMasks:Int = Current Shl ChildShift
			
			If (ChildMasks & $8000) And MinT <= MaxT Then
				If MaxTC*RaySizeCoef + RaySizeBias >= Scale_Exp2 Then Exit
				
				Local MaxTV:Float = Min(MaxT, MaxTC)
				Local Half:Float  = Scale_Exp2*0.5
				Local CenterTX:Float = Half*dTx + CornerTX
				Local CenterTY:Float = Half*dTy + CornerTY
				Local CenterTZ:Float = Half*dTz + CornerTZ
				
				If MinT <= MaxTV Then
					Local ChildOffset:Int = Current Shr 17
					If Current & $10000 Then ChildOffset = FarPointers[ChildOffset]
					
					If (Not (ChildMasks & $0080)) Then
						Normal = Data[ChildOffset + Parent + BitCount[((ChildMasks Shr (8 + ChildShift)) Shl ChildShift) & %1111111]]
						
						Exit'Or Scale = MinLevel Then Exit
					EndIf
					
					If MaxTC < H Then
						Stack[Scale, 0] = Parent
						Stack[Scale, 1] = FloatToInt(MaxT)
					EndIf
					
					H = MaxTC
					
					ChildOffset :+ BitCount[ChildMasks & %1111111]
					Parent :+ ChildOffset
					
					IDx = 0
					Scale :- 1
					Scale_Exp2 = Half
					
					If CenterTX > MinT Then IDx = IDx ~ %001; PosX :+ Scale_Exp2
					If CenterTY > MinT Then IDx = IDx ~ %010; PosY :+ Scale_Exp2
					If CenterTZ > MinT Then IDx = IDx ~ %100; PosZ :+ Scale_Exp2
					
					MaxT = MaxTV
					Current = 0
					Continue
				EndIf
			EndIf
			
			Local StepMask:Int
			If CornerTX <= MaxTC Then StepMask = StepMask ~ %001; PosX :- Scale_Exp2
			If CornerTY <= MaxTC Then StepMask = StepMask ~ %010; PosY :- Scale_Exp2
			If CornerTZ <= MaxTC Then StepMask = StepMask ~ %100; PosZ :- Scale_Exp2
			
			MinT = MaxTC
			IDx = IDx ~ StepMask
			
			If IDx & StepMask Then
				Local DifferingBits:Int
				
				If StepMask & %001 Then DifferingBits = DifferingBits | (FloatToInt(PosX) ~ FloatToInt(PosX + Scale_Exp2))
				If StepMask & %010 Then DifferingBits = DifferingBits | (FloatToInt(PosY) ~ FloatToInt(PosY + Scale_Exp2))
				If StepMask & %100 Then DifferingBits = DifferingBits | (FloatToInt(PosZ) ~ FloatToInt(PosZ + Scale_Exp2))
				Scale = (FloatToInt(Float(DifferingBits)) Shr 23) - 127
				Scale_Exp2 = IntToFloat((Scale - MAX_SCALE + 127) Shl 23)
				
				Parent = Stack[Scale, 0]
				MaxT = IntToFloat(Stack[Scale, 1])
				
				Local ShX:Int = FloatToInt(PosX) Shr Scale
				Local ShY:Int = FloatToInt(PosY) Shr Scale
				Local ShZ:Int = FloatToInt(PosZ) Shr Scale
				PosX = IntToFloat(ShX Shl Scale)
				PosY = IntToFloat(ShY Shl Scale)
				PosZ = IntToFloat(ShZ Shl Scale)
				IDx = (ShX & 1) | ((ShY & 1) Shl 1) | ((ShZ & 1) Shl 2)
				
				H = 0.0
				Current = 0
			EndIf
		Wend
		
		If Scale >= MAX_SCALE Then Return MISS
		
		If (OctantMask & %001) = 0 Then PosX = 3.0 - Scale_Exp2 - PosX
		If (OctantMask & %010) = 0 Then PosY = 3.0 - Scale_Exp2 - PosY
		If (OctantMask & %100) = 0 Then PosZ = 3.0 - Scale_Exp2 - PosZ
		
		Return MinT
	End Method
	
	Method CubeContainsVoxels:Int(Voxels:Byte[,,], StartX:Int, StartY:Int, StartZ:Int, Size:Int)
		If Size = 1 Then Return Voxels[StartX, StartY, StartZ] = 255
		
		For Local X:Int = StartX Until StartX + Size
			For Local Y:Int = StartY Until StartY + Size
				For Local Z:Int = StartZ Until StartZ + Size
					If Voxels[X, Y, Z] = 255 Then Return True
				Next
			Next
		Next
	End Method
	
	Method BuildOctree:Int(Voxels:Byte[,,], Normals:Int[,,], X:Int, Y:Int, Z:Int, Size:Int, DescriptorIndex:Int, WriteIndex:Int)
		Local HalfSize:Int = Size Shr 1
		
		Local PosX:Int[] = [X + HalfSize, X, X + HalfSize, X, X + HalfSize, X, X + HalfSize, X]
		Local PosY:Int[] = [Y + HalfSize, Y + HalfSize, Y, Y, Y + HalfSize, Y + HalfSize, Y, Y]
		Local PosZ:Int[] = [Z + HalfSize, Z + HalfSize, Z + HalfSize, Z + HalfSize, Z, Z, Z, Z]
		
		Local Children:Byte, Leafs:Byte, ChildIndex:Int, ChildCount:Int
		
		For Local I:Int = 0 Until 8
			If CubeContainsVoxels(Voxels, PosX[I], PosY[I], PosZ[I], HalfSize) Then Children :| %10000000 Shr I
		Next
		
		'If HalfSize > 16 Then Print Size + ": " + Children
		
		Leafs = Children
		
		Local ChildWriteIndex:Int
		
		If HalfSize = 1 Then
			Leafs = 0
			
			ChildWriteIndex = WriteIndex - 1
			
			EnsureSpace(Data, DataSize, WriteIndex*4, 8*4)
			
			For Local I:Int = 0 Until 8
				If Children & (1 Shl I) Then
					ChildWriteIndex :+ 1
					
					Data[ChildWriteIndex] = Normals[PosX[7 - I], PosY[7 - I], PosZ[7 - I]]
				EndIf
			Next
		Else
			ChildWriteIndex = WriteIndex + BitCount[Children]
			
			Local ChildrenWritten:Int
			For Local I:Int = 0 Until 8
				If Children & (1 Shl I) Then
					ChildWriteIndex = BuildOctree(Voxels, Normals, PosX[7 - I], PosY[7 - I], PosZ[7 - I], HalfSize, WriteIndex + ChildrenWritten, ChildWriteIndex) + 1
					
					ChildrenWritten :+ 1
				EndIf
			Next
		EndIf
		
		EnsureSpace(Data, DataSize, DescriptorIndex*4, 4)
		
		ChildIndex = WriteIndex - DescriptorIndex
		
		If ChildIndex > 32767 Then
			EnsureSpace(FarPointers, FarPointerSize, FarPointerCount*4, 4)
			
			FarPointers[FarPointerCount] = ChildIndex
			
			Print "FarPointer[" + FarPointerCount + "] = " + ChildIndex
			
			Data[DescriptorIndex] = (FarPointerCount Shl 17) | $10000 | (Children Shl 8) | Leafs
			
			FarPointerCount :+ 1
		Else
			Data[DescriptorIndex] = (ChildIndex Shl 17) | (Children Shl 8) | Leafs
		EndIf
		
		Return ChildWriteIndex
	End Method
	
	Method EnsureSpace(Buffer:Int Ptr Var, Size:Int Var, Index:Int, WriteSize:Int)
		If Index + WriteSize >= Size Then
			Buffer = Int Ptr MemExtend(Buffer, Size, Index + WriteSize + ALLOC_STEP)
			Size   = Index + WriteSize + ALLOC_STEP
			
			'Print "Resizing data to " + ((Index + WriteSize + ALLOC_STEP)/1024) + "kb"
		EndIf
	End Method
	
	Method Render(X:Float, Y:Float, Z:Float, Size:Float = 1.0, Level:Int = 1, Index:Int = 0)
		glColor3f(1.0 - Level/Float(MAX_SCALE), 1.0 - Level/Float(MAX_SCALE), 1.0 - Level/Float(MAX_SCALE))
		WireCube(X, Y, Z, Size)
		
		Local HalfSize:Float = Size*0.5
		
		Local PosX:Float[] = [X + HalfSize, X, X + HalfSize, X, X + HalfSize, X, X + HalfSize, X]
		Local PosY:Float[] = [Y + HalfSize, Y + HalfSize, Y, Y, Y + HalfSize, Y + HalfSize, Y, Y]
		Local PosZ:Float[] = [Z + HalfSize, Z + HalfSize, Z + HalfSize, Z + HalfSize, Z, Z, Z, Z]
		
		Local Descriptor:Int = Data[Index]
		Local Far:Int        = Descriptor & $10000
		Local ChildIndex:Int = Descriptor Shr 17
		Local Children:Int   = (Descriptor Shr 8) & $FF
		Local Leafs:Int      = Descriptor & $FF
		
		If Far Then ChildIndex = FarPointers[ChildIndex]
		
		Local ChildrenDrawn:Int
		For Local I:Int = 0 Until 8
			If Children & (1 Shl I) And Leafs & (1 Shl I) Then
				Render(PosX[7 - I], PosY[7 - I], PosZ[7 - I], HalfSize, Level + 1, Index + ChildIndex + ChildrenDrawn)
				
				ChildrenDrawn :+ 1
			EndIf
		Next
	End Method
	
	Method Save(File:String)
		Local Outstream:TStream = WriteFile(File)
		
		If Outstream Then
			Outstream.WriteInt(FarPointerCount)
			Outstream.WriteBytes(FarPointers, FarPointerCount*4)
			Outstream.WriteInt(DataSize)
			Outstream.WriteBytes(Data, DataSize)
			
			Outstream.Close()
		EndIf
	End Method
	
	Method Load:TVoxelOctree(File:String)
		Local Instream:TStream = ReadFile(File)
		
		If Instream Then
			If FarPointers Then MemFree(FarPointers)
			If Data Then MemFree(Data)
			
			FarPointerCount = InStream.ReadInt()
			FarPointerSize  = FarPointerCount*4
			FarPointers     = Int Ptr MemAlloc(FarPointerSize)
			
			Print FarPointerCount
			
			Instream.ReadBytes(FarPointers, FarPointerSize)
			
			DataSize = InStream.ReadInt()
			Data     = Int Ptr MemAlloc(DataSize)
			
			InStream.ReadBytes(Data, DataSize)
			
			InStream.Close()
			
			Return Self
		EndIf
	End Method
	
	Method Create:TVoxelOctree(Voxels:Byte[,,], Normals:Int[,,])
		If FarPointers Then MemFree FarPointers
		If Data Then MemFree(Data)
		
		Data     = Int Ptr MemAlloc(INIT_DATA_SIZE)
		DataSize = INIT_DATA_SIZE
		
		FarPointers    = Int Ptr MemAlloc(INIT_POINTER_SIZE)
		FarPointerSize = INIT_POINTER_SIZE
		
		Local LastIndex:Int = BuildOctree(Voxels, Normals, 0, 0, 0, Voxels.Dimensions()[0], 0, 1)
		Print "DataSize: " + ((DataSize)/1024) + "kb Data used: " + ((LastIndex*4)/1024) + "kb FarPointers: " + FarPointerCount
		
		Return Self
	End Method
	
	Function FloatToInt:Int(In:Float)
		Return (Int Ptr (Varptr In))[0]
	End Function
	
	Function IntToFloat:Float(In:Int)
		Return (Float Ptr (Varptr In))[0]
	End Function
	
	Function WireCube(X:Float, Y:Float, Z:Float, S:Float)
		glBegin(GL_LINE_STRIP)
			glVertex3f(X    , Y    , Z    )
			glVertex3f(X + S, Y    , Z    )
			glVertex3f(X + S, Y + S, Z    )
			glVertex3f(X    , Y + S, Z    )
			glVertex3f(X    , Y    , Z    )
			glVertex3f(X    , Y    , Z + S)
			glVertex3f(X + S, Y    , Z + S)
			glVertex3f(X + S, Y + S, Z + S)
			glVertex3f(X    , Y + S, Z + S)
			glVertex3f(X    , Y    , Z + S)
		glEnd()
		
		glBegin(GL_LINES)
			glVertex3f(X + S, Y    , Z    )
			glVertex3f(X + S, Y    , Z + S)
			
			glVertex3f(X + S, Y + S, Z    )
			glVertex3f(X + S, Y + S, Z + S)
			
			glVertex3f(X    , Y + S, Z    )
			glVertex3f(X    , Y + S, Z + S)
		glEnd()
	End Function
End Type